#include <avr/io.h>

#include <avr/interrupt.h>

#include <util/delay.h>

#include <stdio.h>
#include <string.h>

// button debouncer parameters
#define BUT_NCHECKS		4
#define BUT_IMASK		0x03

// sampling speed grades
#define S_LIGHT			1
#define S_RIDICULOUS	2
#define S_LUDICROUS		3

// buttons
uint8_t but_index;
uint8_t but_state[BUT_NCHECKS];
uint8_t but_count;
uint8_t but_repeat;

// array of samples
uint8_t samples[200];

// sample acquisition
volatile uint8_t samp_flag;
volatile uint8_t sample_new;
volatile uint8_t sample_old;
uint8_t samp_index;
uint8_t samp_acq_index;

// time division
uint8_t samp_speed_grade;
uint8_t samp_rate;
uint16_t samp_rate_table[13] =
{
		62500, // 5s, fosc/8
		25000, // 2s
		12500, // 1s
		6250, // 0.5s
		20000, // 0.2s, fosc/1
		10000, // 0.1s
		5000, // 50ms
		2000, // 20ms
		1000, // 10ms
		500, // 5ms
		200, // 2ms
		100, // 1ms
		50 // 0.5ms
};

// voltage division
uint8_t volt_setting;

// capture state
uint8_t running;
uint8_t triggered;
uint8_t single_seq;
volatile uint8_t capt_count;

// parameter being editted
uint8_t edit_mode;

// trigger and cursor
uint8_t trig_lvl;
uint8_t x_curs;

// voltage divisions in text
char *volt_text[] =
{
		"5.0V ",
		"2.5V ",
		"1.25V",
		"1.0V ",
		"625mV",
		"500mV",
		"313mV",
		"156mV"
};

// voltage divisions in microvolts
int32_t volt_uv[] =
{
		5000000,
		2500000,
		1250000,
		1000000,
		625000,
		500000,
		312500,
		156250
};

// time divisions in text
char *time_text[] =
{
		"5.0s ",
		"2.0s ",
		"1.0s ",
		"500ms",
		"200ms",
		"100ms",
		"50ms ",
		"20ms ",
		"10ms ",
		"5.0ms",
		"2.0ms",
		"1.0ms",
		"0.5ms"
};

// time divisions in microseconds
int32_t time_us[] =
{
		5000000,
		2000000,
		1000000,
		500000,
		200000,
		100000,
		50000,
		20000,
		10000,
		5000,
		2000,
		1000,
		500
};

// header and footer
char header[26] = ">H:       V:            R\0";
char footer[26] = " Ct:         CV:         \0";

// footer trigger and cursor readouts
char footer_trig[25] = "T:                      \0";
char footer_curs[25] = "Ct:         CV:         \0";

/*
 * INT1: Reads sample from ADC
 */
ISR(INT1_vect)
{
	sample_old = sample_new;
	PORTD &= ~(_BV(PD5));
	sample_new = PINC;
	PORTD |= _BV(PD5);

	samp_flag = 1;
}

/*
 * TIMER2_COMPA: Button debouncer, also generates ~10ms time base
 */
ISR(TIMER2_COMPA_vect, ISR_NOBLOCK)
{
	but_state[but_index++ & BUT_IMASK] = PINA;
	if (capt_count > 0) capt_count--;
	if (but_repeat > 0) but_repeat--;
}

/*
 * TIMER1_COMPA: Starts an ADC conversion
 */
ISR(TIMER1_COMPA_vect, ISR_NAKED)
{
	// Start conversion
	asm volatile(
			"cbi 0x0B, 4"		"\n\t"
			"sbi 0x0B, 4"		"\n\t"
			"reti"
			: :
			);
}

// function prototypes
void verify();
void initialize();

void sendPGA(uint8_t, uint8_t);

void sendVideoCmnd(uint8_t);
void sendVideoSamp(uint8_t);
void sendHeader();
void sendFooter();

void setMode(uint8_t);
void setTriggerReadout(uint8_t);
void setCursorReadout(uint8_t);
void setWindow(uint8_t,uint8_t);

inline void captureLight();
inline void captureRidiculous();
inline void captureLudicrous();

int main()
{
	uint8_t i;
	uint8_t but = 0, last_but = 0, but_flag = 0, rel_flag = 0;

	verify();
	initialize();

	sei();

	sendVideoCmnd(0xA0);
	setMode(0);
	setTriggerReadout(trig_lvl);
	setWindow(samp_rate, volt_setting);

	while (1)
	{
		last_but = but;
		but = 0xFF;
		for (i = 0; i < BUT_NCHECKS; i++)
			but &= ~(but_state[i]);
		but_flag = but & ~last_but;
		rel_flag = ~but & last_but;

		// UP
		if (but_flag & 0x01)
		{
			if (edit_mode == 0 && samp_rate < 12)
			{
				// increase time division
				setWindow(++samp_rate, volt_setting);
			}
			else if (edit_mode == 1 && volt_setting < 7)
			{
				// increase voltage division
				setWindow(samp_rate, ++volt_setting);
			}
			else if (edit_mode == 2 && x_curs < 199)
			{
				// increase cursor coordinate
				sendVideoCmnd(0xA3);
				sendVideoCmnd(++x_curs);
				but_repeat = 50;
			}
			else if (edit_mode == 3 && trig_lvl < 255)
			{
				// increase trigger level
				sendVideoCmnd(0xA4);
				sendVideoSamp(++trig_lvl);
				setTriggerReadout(trig_lvl);
				sendFooter();
				but_repeat = 50;
			}
		}
		else if (but & 0x01 && but_repeat == 0)
		{
			if (edit_mode == 2 && x_curs < 199)
			{
				// continue increasing cursor coordinate
				sendVideoCmnd(0xA3);
				sendVideoCmnd(++x_curs);
				but_repeat = 5;
			}
			else if (edit_mode == 3 && trig_lvl < 255)
			{
				// continue increasing trigger level
				sendVideoCmnd(0xA4);
				sendVideoSamp(++trig_lvl);
				setTriggerReadout(trig_lvl);
				sendFooter();
				but_repeat = 5;
			}
		}
		else if (rel_flag & 0x01)
		{
			if (edit_mode == 2)
			{
				// update cursor readout
				setCursorReadout(x_curs);
				sendFooter();
			}
			else if (edit_mode == 3)
			{
				// erase trigger indicator on video
				sendVideoCmnd(0xA4);
				sendVideoCmnd(255);
			}
		}

		// DOWN
		if (but_flag & 0x02)
		{
			if (edit_mode == 0 && samp_rate > 0)
			{
				// decrease time division
				setWindow(--samp_rate, volt_setting);
			}
			else if (edit_mode == 1 && volt_setting > 0)
			{
				// decrease voltage division
				setWindow(samp_rate, --volt_setting);
			}
			else if (edit_mode == 2 && x_curs > 0)
			{
				// decrease cursor coordinate
				sendVideoCmnd(0xA3);
				sendVideoCmnd(--x_curs);
				but_repeat = 50;
			}
			else if (edit_mode == 3 && trig_lvl > 0)
			{
				// decrease trigger level
				sendVideoCmnd(0xA4);
				sendVideoSamp(--trig_lvl);
				setTriggerReadout(trig_lvl);
				sendFooter();
				but_repeat = 50;
			}
		}
		else if (but & 0x02 && but_repeat == 0)
		{
			if (edit_mode == 2 && x_curs > 0)
			{
				// continue decreasing cursor coordinate
				sendVideoCmnd(0xA3);
				sendVideoCmnd(--x_curs);
				but_repeat = 5;
			}
			else if (edit_mode == 3 && trig_lvl > 0)
			{
				// continue decreasing trigger level
				sendVideoCmnd(0xA4);
				sendVideoSamp(--trig_lvl);
				setTriggerReadout(trig_lvl);
				sendFooter();
				but_repeat = 5;
			}
		}
		else if (rel_flag & 0x02)
		{
			if (edit_mode == 2)
			{
				// update cursor readout
				setCursorReadout(x_curs);
				sendFooter();
			}
			else if (edit_mode == 3)
			{
				// erase trigger indicator on video
				sendVideoCmnd(0xA4);
				sendVideoCmnd(255);
			}
		}

		// CRS
		if (but_flag & 0x04)
		{
			if (!running)
				setMode(2);
		}

		// VOLT
		if (but_flag & 0x08)
		{
			setMode(1);
		}

		// TIME
		if (but_flag & 0x10)
		{
			setMode(0);
		}

		// TRIG
		if (but_flag & 0x20)
		{
			setMode(3);
		}

		// SEQ
		if (but_flag & 0x40)
		{
			if (!running)
			{
				// write SEQ in header
				header[20] = 'S';
				header[21] = 'E';
				header[22] = 'Q';
				setWindow(samp_rate, volt_setting);
				single_seq = 1;
			}
		}

		// RUN
		if (but_flag & 0x80)
		{
			if (running)
			{
				// stop running
				running = 0;
				header[24] = 'S';
				sendHeader();
				sendFooter();
			}
			else
			{
				// start running
				running = 1;
				header[24] = 'R';
				sendVideoCmnd(0xA3);
				sendVideoCmnd(255);
				if (edit_mode == 2)
					setMode(3);
				if (single_seq)
				{
					header[20] = ' ';
					header[21] = ' ';
					header[22] = ' ';
					single_seq = 0;
				}
				setTriggerReadout(trig_lvl);
				setWindow(samp_rate, volt_setting);
			}
		}

		// capture data
		if (running || single_seq)
		{
			if (samp_speed_grade == S_LIGHT)
				captureLight();
			else if (samp_speed_grade == S_RIDICULOUS)
				captureRidiculous();
			else if (samp_speed_grade == S_LUDICROUS)
				captureLudicrous();

			// single sequence complete, erase SEQ
			if (!running && !single_seq)
			{
				header[20] = ' ';
				header[21] = ' ';
				header[22] = ' ';
				sendHeader();
			}
		}
	}

	return 0;
}

/*
 * SENDPGA: Send instruction inst and value val to PGA
 */
void sendPGA(uint8_t inst, uint8_t val)
{
	// set CS low
	PORTB &= ~(_BV(PB0));

	// send instruction, wait for transmission to complete
	SPDR = inst;
	while (!(SPSR & 0x80)) ;

	// send value, wait for transmission to complete
	SPDR = val;
	while (!(SPSR & 0x80)) ;

	// set CS high
	PORTB |= _BV(PB0);
}

/*
 * SENDVIDEO: Send a byte to the video microcontroller
 */
inline void sendVideo(uint8_t c)
{
	// wait for ACC high
	while (!(PINB & _BV(PINB2))) ;

	// set REQ low
	PORTB &= ~(_BV(PB3));

	// wait for ACC low
	while (PINB & _BV(PINB2)) ;

	// set CS low
	PORTB &= ~(_BV(PB4));

	// send byte, wait for transmission to complete
	SPDR = c;
	while (!(SPSR & 0x80)) ;

	// set CS high
	PORTB |= _BV(PB4);

	// set REQ high
	PORTB |= _BV(PB3);
}

/*
 * SENDVIDEOCMND: Send a command to the video microcontroller
 */
void sendVideoCmnd(uint8_t c)
{
	sendVideo(c);
}

/*
 * SENDVIDEOSAMP: Send a sample to the video microcontroller
 */
void sendVideoSamp(uint8_t s)
{
	sendVideo(127-(uint8_t)((((int8_t)(s-128))>>1) + 64));
}

/*
 * SENDHEADER: Send the header string to the video microcontroller
 */
void sendHeader()
{
	uint8_t i;

	sendVideoCmnd(0xA1);
	for (i = 0; i < 25; i++)
		sendVideoCmnd(header[i]);
	sendVideoCmnd(0xA0);
}

/*
 * SENDFOOTER: Send the footer string to the video microcontroller
 */
void sendFooter()
{
	uint8_t i;

	sendVideoCmnd(0xA2);
	for (i = 0; i < 25; i++)
		sendVideoCmnd(footer[i]);
	sendVideoCmnd(0xA0);
}

/*
 * SETMODE: Sets the current edit mode to m
 */
void setMode(uint8_t m)
{
	// erase mode indicators
	header[0] = ' ';
	header[9] = ' ';
	footer[0] = ' ';

	// place mode indicator appropriately
	if (m == 0)
		header[0] = '>';
	else if (m == 1)
		header[9] = '>';
	else if (m == 2)
	{
		setCursorReadout(x_curs);
		footer[0] = '>';
	}
	else if (m == 3)
	{
		setTriggerReadout(trig_lvl);
		footer[0] = '>';
	}

	// send header & footer
	sendHeader();
	sendFooter();

	// turn on cursor if entering cursor mode, turn of if exiting
	if (edit_mode == 2 && m != 2)
	{
		sendVideoCmnd(0xA3);
		sendVideoCmnd(255);
	}
	else if (m == 2)
	{
		sendVideoCmnd(0xA3);
		sendVideoCmnd(x_curs);
	}

	// set current mode
	edit_mode = m;
}

/*
 * SETTRIGGERREADOUT: Set the footer string to have the trigger level t
 */
void setTriggerReadout(uint8_t t)
{
	int32_t y = (int32_t)((int8_t)(t-128));
	int32_t v;
	int16_t l,r;
	uint8_t neg = 0;

	strcpy(footer+1, footer_trig);

	v = (volt_uv[volt_setting]*y)>>8;
	if (v < 0)
	{
		neg = 1;
		v = -v;
	}

	if (volt_setting < 4)
	{
		l = (int16_t)(v/1000000);
		r = (int16_t)((v%1000000)/1000);
		snprintf(footer+4, 8, "%d.%03dV ", l, r);
	}
	else
	{
		l = (int16_t)(v/1000);
		r = (int16_t)((v%1000)/100);
		snprintf(footer+4, 8, "%3d.%dmV", l, r);
	}

	footer[3] = neg ? '-' : ' ';
}

/*
 * SETCURSORREADOUT: Set the footer string to have the cursor data for
 * 		position x
 */
void setCursorReadout(uint8_t x)
{
	int32_t y = (int32_t)((int8_t)(samples[x]-128));//<<8;
	int32_t v;
	int16_t l,r;
	uint8_t neg = 0;

	strcpy(footer+1, footer_curs);

	v = (time_us[samp_rate]*x)/200;

	if (samp_rate < 3)
	{
		l = (int16_t)(v/1000000);
		r = (int16_t)((v%1000000)/1000);
		snprintf(footer+4, 9, "%d.%03ds ", l, r);
	}
	else if (samp_rate < 6)
	{
		l = (int16_t)(v/1000);
		r = (int16_t)((v%1000)/100);
		snprintf(footer+4, 9, "%3d.%dms", l, r);
	}
	else if (samp_rate < 9)
	{
		l = (int16_t)(v/1000);
		r = (int16_t)((v%1000)/10);
		snprintf(footer+4, 9, "%2d.%02dms", l, r);
	}
	else if (samp_rate < 12)
	{
		l = (int16_t)(v/1000);
		r = (int16_t)(v%1000);
		snprintf(footer+4, 9, "%d.%03dms", l, r);
	}
	else
	{
		l = (int16_t)(v);
		snprintf(footer+4, 9, "%3dus  ", l);
	}

	v = (volt_uv[volt_setting]*y)>>8;
	if (v < 0)
	{
		neg = 1;
		v = -v;
	}

	if (volt_setting < 4)
	{
		l = (int16_t)(v/1000000);
		r = (int16_t)((v%1000000)/1000);
		snprintf(footer+17, 8, "%d.%03dV ", l, r);
	}
	else
	{
		l = (int16_t)(v/1000);
		r = (int16_t)((v%1000)/100);
		snprintf(footer+17, 8, "%3d.%dmV", l, r);
	}

	footer[16] = neg ? '-' : ' ';
}

/*
 * SETWINDOW: Set the graphing window, rate is sample rate, volt is voltage
 * 		division
 */
void setWindow(uint8_t rate, uint8_t volt)
{
	// turn off and reset timer 1
	TCCR1B &= ~(_BV(CS12) | _BV(CS11) | _BV(CS10));
	TIMSK1 &= ~(_BV(OCIE1A));
	TIFR1 |= _BV(OCF1A);
	TCNT1 = 0;

	// set sample rate
	OCR1A = samp_rate_table[rate];

	// disable external interrupt 1
	EIMSK &= ~(_BV(INT1));
	EIFR |= _BV(INTF1);

	// re-initialize various variables
	triggered = 0;
	samp_flag = 0;
	samp_acq_index = 0;

	// update header
	strcpy(header+3, time_text[rate]);
	strcpy(header+12, volt_text[volt]);

	// send header and footer to video
	sendHeader();
	sendFooter();

	// send gain to PGA
	sendPGA(0x40, volt);

	if (rate < 4)
	{
		// Light speed!
		sendVideoCmnd(0x90);
		samp_index = 0;
		samp_speed_grade = S_LIGHT;
		TCCR1B |= _BV(CS11);	// turn on timer 1 with clock fosc/8
		TIMSK1 |= _BV(OCIE1A);	// turn on OC1A interrupt
		EIMSK |= _BV(INT1);		// turn on external interrupt 1
	}
	else if (rate < 9)
	{
		// Ridiculous speed!
		sendVideoCmnd(0x80);
		samp_index = 200;
		sample_old = 255;
		samp_speed_grade = S_RIDICULOUS;
		TCCR1B |= _BV(CS10);	// turn on timer 1 with clock fosc/1
		TIMSK1 |= _BV(OCIE1A);	// turn on OC1A interrupt
		EIMSK |= _BV(INT1);		// turn on external interrupt 1
	}
	else
	{
		// LUDICROUS SPEED!!!
		sendVideoCmnd(0x80);
		samp_index = 200;
		samp_speed_grade = S_LUDICROUS;
		// leave timer 1 off
		// leave OC1A interrupt off
		// leave external interrupt 1 off
		capt_count = 0; // start sampling immediately
	}
}

/*
 * CAPTURELIGHT: Capture samples at the slow speed grade, immediately
 * 		display samples
 */
inline void captureLight()
{
	if (samp_flag)
	{
		samp_flag = 0;

		// store sample
		samples[samp_index++] = sample_new;

		// send sample
		sendVideoSamp(sample_new);

		if (samp_index == 200)
		{
			sendVideoCmnd(0x80);
			samp_index = 0;
			single_seq = 0;
		}
	}
}

/*
 * CAPTURERIDICULOUS: Capture samples at the medium speed grade, capture
 * 		200 samples after trigger and display after all captured
 */
inline void captureRidiculous()
{
	uint8_t curr, prev;

	if (samp_flag && capt_count == 0)
	{
		samp_flag = 0;

		curr = sample_new;
		prev = sample_old;

		// check for trigger
		if (prev < trig_lvl && curr >= trig_lvl)
			triggered = 1;

		if (triggered)
		{
			// store sample
			samples[samp_acq_index++] = curr;

			if (samp_acq_index == 200)
			{
				samp_acq_index = 0;
				triggered = 0;

				// start sample transmission
				sendVideoCmnd(0x80);
				samp_index = 0;

				// wait for 500ms before next read cycle
				capt_count = 50;
			}
		}
	}

	if (samp_index < 200)
	{
		// send sample
		sendVideoSamp(samples[samp_index++]);
		if (samp_index == 200)
			single_seq = 0;
	}
}

/*
 * CAPTURELUDICROUS: Capture samples at the medium speed grade, capture
 * 		200 samples after trigger and display after all captured
 */
inline void captureLudicrous()
{
	uint8_t curr_samp = 255, prev_samp;
	uint8_t* samp_ptr;
	uint8_t timeout = 200;

	if (capt_count == 0)
	{
		PIND = _BV(PIND7);

		samp_ptr = samples;
		triggered = 0;

		// temporarily shut off button debouncer
		TCCR2B &= ~(_BV(CS22) | _BV(CS21) | _BV(CS20));

		EIFR |= _BV(INTF1);		// make sure INT1 flag is clear
		TCNT1 = 0;				// reset counter 1
		TIFR1 |= _BV(OCF1A);	// make sure OC1A flag is clear
		TCCR1B |= _BV(CS10);	// start timer 1
		TIMSK1 |= _BV(OCIE1A);	// turn on OC1A interrupt

		// wait for trigger, then acquire 200 samples
		while (samp_ptr < samples+200 && timeout > 0)
		{
			prev_samp = curr_samp;
			while (!(EIFR & _BV(INTF1))) ;
			EIFR |= _BV(INTF1);

			PORTD &= ~(_BV(PD5));
			curr_samp = PINC;
			PORTD |= _BV(PD5);

			if (prev_samp < trig_lvl && curr_samp >= trig_lvl)
				triggered = 1;

			if (triggered)
				*(samp_ptr++) = curr_samp;
			else
				timeout--;
		}

		TIMSK1 &= ~(_BV(OCIE1A));	// turn off OC1A interrupt
		TCCR1B &= ~(_BV(CS10));		// stop timer 1

		// restart button debouncer
		TCCR2B |= _BV(CS22) | _BV(CS21) | _BV(CS20);

		if (timeout > 0)
		{
			// start sample transmission
			sendVideoCmnd(0x80);
			samp_index = 0;

			// wait for 500ms before next read cycle
			capt_count = 50;
		}
		else
		{
			// timed out, wait for 500ms before trying again
			capt_count = 50;
		}
	}

	if (samp_index < 200)
	{
		// send sample
		sendVideoSamp(samples[samp_index++]);
		if (samp_index == 200)
			single_seq = 0;
	}
}

/*
 * INITIALIZE: Set up the microcontroller registers and initialize global
 * 		variables.
 */
void initialize()
{
	uint8_t i;

	// PA7: RUN			in
	// PA6: SEQ			in
	// PA5: TRIG		in
	// PA4: TIME		in
	// PA3: VOLT		in
	// PA2: CRS			in
	// PA1: DOWN		in
	// PA0: UP			in
	DDRA = 0;
	PORTA = 0xFF; // internal pull-ups

	// PB7: SCK			out
	// PB6: MISO		in
	// PB5: MOSI		out
	// PB4: VIDEO CS	out
	// PB3: VIDEO REQ	out
	// PB2: VIDEO ACC	in
	// PB1: (unused)
	// PB0: PGA CS		out
	PORTB = _BV(PB4) | _BV(PB3) | _BV(PB0);
	DDRB = _BV(DDB7) | _BV(DDB5) | _BV(DDB4) | _BV(DDB3) | _BV(DDB0);

	// PC7-0: ADC data	in
	DDRC = 0;

	// PD7: LED			out
	// PD6: (unused)
	// PD5: ADC RD		out
	// PD4: ADC CONVST	out
	// PD3: ADC EOC		in
	// PD2: EXT TRIG	in
	// PD1: RS232 TX
	// PD0: RS232 RX
	PORTD = _BV(PD5) | _BV(PD4);
	DDRD = _BV(DDD7) | _BV(DDD5) | _BV(DDD4);

	_delay_ms(5);

	// INT1: ADC EOC (falling edge)
	EICRA = _BV(ISC11);
	// don't enable external interrupt 1 yet

	// SPI: MSB first, master, mode 00, fosc/4 (5 MHz)
	SPCR = _BV(SPE) | _BV(MSTR);

	// TIMER 1 (sample rate): OC1* disconnected, CTC mode
	TCCR1B = _BV(WGM12);	// leave it off for now

	// TIMER 2 (buttons): OC2* disconnected, CTC mode, fosc/1024 (~19.5kHz),
	//		OC2A interrupt enabled
	TCCR2A = _BV(WGM21);
	TCCR2B = _BV(CS22) | _BV(CS21) | _BV(CS20);
	OCR2A = 195; // ~10 ms time base
	TIMSK2 = _BV(OCIE2A);

	// Button state
	for (i = 0; i < BUT_NCHECKS; i++)
		but_state[i] = 0xFF;
	but_index = 0;

	// Samples
	samp_flag = 0;
	samp_index = 0;
	samp_rate = 0;
	samp_acq_index = 0;

	running = 1;

	triggered = 0;

	single_seq = 0;

	// trigger level at 0V
	trig_lvl = 128;

	edit_mode = 0;
	volt_setting = 0;

	// cursor in middle
	x_curs = 100;
}

/*
 * VERIFY: Check to make sure the code is running on the correct
 * 		microcontroller. If not, flash a warning on the LED forever.
 */
void verify()
{
	uint8_t i;

	// if pin 4 on port D is high, everything's fine!
	if (PIND & _BV(PIND4))
		return;

	// we're executing on the wrong microcontroller!
	DDRD |= _BV(DDD7);
	while (1)
	{
		// flash the LED forever
		PORTD |= _BV(PD7);
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD &= ~(_BV(PD7));
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD |= _BV(PD7);
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD &= ~(_BV(PD7));
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD |= _BV(PD7);
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD &= ~(_BV(PD7));
		for (i = 0; i < 100; i++)
			_delay_ms(10);
	}
}
